from proteome.plugins.unite.main import Plugin
from proteome.plugins.unite.data import UniteSelectAdd, UniteSelectAddAll, UniteProjects, UniteNames

__all__ = ('Plugin', 'UniteSelectAdd', 'UniteSelectAddAll', 'UniteProjects', 'UniteNames')



from ribosome import RootMachine
from ribosome.nvim import HasNvim
from ribosome.machine import ModularMachine, Transitions
from ribosome.machine.base import MachineBase

from proteome.nvim import NvimFacade
from proteome.logging import Logging


class ProteomeComponent(ModularMachine, HasNvim, Logging):

    def __init__(self, vim: NvimFacade, parent=None, title=None) -> None:
        MachineBase.__init__(self, parent, title=title)
        HasNvim.__init__(self, vim)


class ProteomeState(RootMachine, Logging):

    @property
    def title(self):
        return 'proteome'


class ProteomeTransitions(Transitions, HasNvim):

    def __init__(self, machine, *a, **kw):
        Transitions.__init__(self, machine, *a, **kw)
        HasNvim.__init__(self, machine.vim)

__all__ = ('ProteomeComponent', 'ProteomeState', 'ProteomeTransitions')



import amino.logging
from ribosome.logging import ribosome_logger

from amino.lazy import lazy


log = proteome_root_logger = ribosome_logger('proteome')


def proteome_logger(name: str):
    return proteome_root_logger.getChild(name)


class Logging(amino.logging.Logging):

    @lazy
    def _log(self) -> amino.logging.Logger:  # type: ignore
        return proteome_logger(self.__class__.__name__)

__all__ = ('proteome_logger', 'Logging')



import re
from pathlib import Path

from amino import List, Map, Empty, may, _, L
from amino.lazy import lazy

from ribosome.machine import handle, may_handle, Error, Info, Nop
from ribosome.process import JobClient
from ribosome.machine.transition import may_fallback

from proteome.state import ProteomeComponent, ProteomeTransitions
from proteome.project import Project, mkpath
from proteome.git import Git
from proteome.plugins.core.message import (StageI, StageIV, Add, RemoveByIdent, Create, Next, Prev, SetProject,
                                           SetProjectIdent, SetProjectIndex, SwitchRoot, Added, Removed, ProjectChanged,
                                           BufEnter, Initialized, MainAdded, Show, AddByParams, CloneRepo, StageII,
                                           StageIII)


@may
def valid_index(i, total):
    if i >= 0 and i < total:
        return i
    elif i < 0 and -i <= total:
        return total + i


class CoreTransitions(ProteomeTransitions):

    def _no_such_ident(self, ident: str, params):
        return 'no project found matching \'{}\''.format(ident)

    @may_handle(StageI)
    def stage_1(self):
        main = self.data.analyzer(self.vim).main
        return Add(main), MainAdded().pub  # type: ignore

    @may_fallback(StageII)
    def stage_2(self):
        pass

    @may_fallback(StageIII)
    def stage_3(self):
        pass

    @may_handle(StageIV)
    def stage_4(self):
        return BufEnter(self.vim.buffer).pub, Initialized().pub

    @may_handle(Initialized)
    def initialized(self):
        return self.data.set(initialized=True), SwitchRoot(False)

    @may_handle(MainAdded)
    def main_added(self):
        self.data.main.foreach(self._setup_main)

    def _setup_main(self, pro: Project):
        self.vim.vars.set_p('main_name', pro.name)
        self.vim.vars.set_p('main_ident', pro.ident)
        self.vim.vars.set_p('main_type', pro.tpe | 'none')
        self.vim.vars.set_p('main_types', pro.all_types)

    @may_handle(AddByParams)
    def add_by_params(self):
        options = Map(self.msg.options)
        ident = self.msg.ident
        return (
            (options.get('root') /
             mkpath //
             L(self.data.loader.from_params)(ident, _, params=options))
            .or_else(
                self.data.loader.by_ident(ident)
                .or_else(self.data.loader.resolve_ident(ident, options, self.data.main_type))
            ) /
            Add.from_msg(self.msg) |
            Error(self._no_such_ident(ident, options))
        )

    @may_handle(Add)
    def add(self):
        if self.msg.project not in self.data:
            return self.data.add(self.msg.project), Added.from_msg(self.msg)(self.msg.project).pub

    @may_handle(Added)
    def added(self):
        self.vim.vars.set_p('added_project', self.msg.project.json)
        self.vim.vars.set_p('projects', self.data.projects.json)
        self.vim.pautocmd('Added')
        if self.data.initialized and not self.msg.bang:
            return SetProjectIndex(-1)

    @handle(RemoveByIdent)
    def remove_by_ident(self):
        id = self.msg.ident
        target = self.data.projects.index_of_ident(id) | float('nan')
        cur = self.data.current_index
        switch = SetProjectIndex(0) if target == cur else Nop()
        data = self.data.set_index(cur - 1) if target < cur else self.data
        return self.data.project(id)\
            .map(lambda a: (data - a, Removed(a).pub, switch))

    @may_handle(Removed)
    def removed(self):
        self.vim.vars.set_p('projects', self.data.projects.json)
        self.vim.pautocmd('Removed')
        return Info('Removed project {}'.format(self.msg.project.ident))

    @may_handle(Create)
    def create(self):
        return self.data + Project.of(self.msg.name, Path(self.msg.root))

    @may_handle(Show)
    def show(self):
        lines = self.data.projects.show(List.wrap(self.msg.names))
        header = List('Projects:')  # type: List[str]
        return Info('\n'.join(header + lines))

    @may_handle(SetProject)
    def set_project(self):
        if isinstance(self.msg.ident, str):
            if self.msg.ident in self.data:
                return SetProjectIdent(self.msg.ident)
            elif self.msg.ident.isdigit():
                return SetProjectIndex(int(self.msg.ident))
            else:
                err = '\'{}\' is not a valid project identifier'
                return Error(err.format(self.msg.ident))

    @handle(SetProjectIndex)
    def set_project_index(self):
        return (
            valid_index(self.msg.index, self.data.project_count) /
            self.data.set_index /
            (lambda a: (a, SwitchRoot()))
        )

    @handle(SetProjectIdent)
    def set_project_ident(self):
        return self.data.set_index_by_ident(self.msg.ident)\
            .map(lambda a: (a, SwitchRoot()))

    @handle(SwitchRoot)
    def switch_root(self):
        def go(pro: Project):
            self.vim.switch_root(pro.root)  # type: ignore
            pc = ProjectChanged(pro)
            info = 'switched root to {}'
            return ((pc, Info(info.format(pro.ident))) if self.msg.notify
                    else pc)
        return self.data.current.map(go) if self.data.initialized else Empty()

    @may_handle(ProjectChanged)
    def project_changed(self):
        self.vim.vars.set_p('active', self.msg.project.json)

    @may_handle(Next)
    def next(self):
        return self.data.inc(1), SwitchRoot()

    @may_handle(Prev)
    def prev(self):
        return self.data.inc(-1), SwitchRoot()

    @may_handle(Error)
    def error(self):
        self.log.error(self.msg.message)

    @may_handle(Info)
    def info(self):
        self.log.info(self.msg.message)

    # TODO make configurable (destination dir)
    @handle(CloneRepo)
    def clone_repo(self):
        uri = self.msg.uri
        name = self.msg.options.get('name')\
            .or_else(self._clone_repo_name(uri))
        url = self._clone_url(uri)
        return (
            self.data.main_clone_dir.ap2(name, lambda dir, n: dir / n)
            .to_either('invalid parameter: {}'.format(uri)) /
            L(self._clone_repo)(url, _)
        )

    @may_handle(BufEnter)
    def buf_enter(self):
        ''' dummy handler to prevent error message in tests when ctags
        plugin is not active
        '''
        pass

    def _clone_url(self, uri):
        return uri if uri.startswith('http') else self._github_url(uri)

    def _github_url(self, uri):
        return 'https://github.com/{}'.format(uri)

    def _clone_repo_name(self, uri):
        return (
            List.wrap(uri.split('/'))
            .lift(-1) /
            L(re.sub)('\.git$', '', _)
        )

    @property
    def cloner(self) -> Git:
        return self._cloner  # type: ignore

    @lazy
    def _cloner(self):
        return Git(self.vim)

    async def _clone_repo(self, url: str, target):
        ident = '/'.join(str(target).split('/')[-2:])
        client = JobClient(cwd=Path.home(), name=self.name)
        res = await self.cloner.clone(client, url, target)
        return res.either(
            AddByParams(ident, Map()).pub,
            Error('failed to clone {} to {}'.format(url, target)).pub
        )


class Plugin(ProteomeComponent):
    Transitions = CoreTransitions

__all__ = ('Plugin',)



from ribosome.machine import message, json_message

StageI = message('StageI')
StageII = message('StageII')
StageIII = message('StageIII')
StageIV = message('StageIV')
Add = message('Add', 'project')
RemoveByIdent = message('RemoveByIdent', 'ident')
Create = message('Create', 'name', 'root')
Next = message('Next')
Prev = message('Prev')
SetProject = message('SetProject', 'ident')
SetProjectIdent = message('SetProjectIdent', 'ident')
SetProjectIndex = message('SetProjectIndex', 'index')
SwitchRoot = message('SwitchRoot', opt_fields=(('notify', True),))
Save = message('Save')
Added = message('Added', 'project')
Removed = message('Removed', 'project')
ProjectChanged = message('ProjectChanged', 'project')
BufEnter = message('BufEnter', 'buffer')
Initialized = message('Initialized')
MainAdded = message('MainAdded')
Show = message('Show', varargs='names')
AddByParams = json_message('AddByParams', 'ident')
CloneRepo = json_message('CloneRepo', 'uri')

__all__ = ('StageI', 'StageII', 'StageIII', 'StageIV', 'Add', 'RemoveByIdent', 'Create', 'Next', 'Prev', 'SetProject',
           'SetProjectIdent', 'SetProjectIndex', 'SwitchRoot', 'Save', 'Added', 'Removed', 'ProjectChanged', 'BufEnter',
           'Initialized', 'MainAdded', 'Show', 'AddByParams', 'CloneRepo')



from pathlib import Path

from toolz.itertoolz import cons

from amino import List, Map, Maybe, Empty

from proteome.nvim import NvimFacade
from proteome.env import Env
from proteome.state import ProteomeState
from proteome.project import Projects


class Proteome(ProteomeState):

    def __init__(
            self,
            vim: NvimFacade,
            config_path: Path,
            plugins: List[str],
            bases: List[Path],
            type_bases: Map[Path, List[str]],
            initial_projects: Maybe[Projects]=Empty()
    ) -> None:
        self._config_path = config_path
        self._bases = bases
        self._type_bases = type_bases
        self._initial_projects = initial_projects
        core = 'proteome.plugins.core'
        ProteomeState.__init__(self, vim, List.wrap(cons(core, plugins)))

    @property
    def init(self):
        return Env(  # type: ignore
            config_path=self._config_path,
            bases=self._bases,
            type_bases=self._type_bases,
            projects=self._initial_projects | Projects()
        )

    @property
    def projects(self):
        return self.env.projects

__all__ = ('Proteome',)



from ribosome.machine import may_handle

from amino import List, Map

from proteome.state import ProteomeComponent, ProteomeTransitions
from proteome.plugins.unite.data import (UniteSelectAdd, UniteSelectAddAll,
                                         UniteProjects, UniteSource, UniteKind,
                                         UniteNames, UniteMessage)


class UniteTransitions(ProteomeTransitions):

    def unite_cmd(self, cmd):
        args = ' '.join(self.msg.unite_args)
        self.vim.cmd('Unite {} {}'.format(cmd, args))

    @may_handle(UniteSelectAdd)
    def select_add(self):
        self.unite_cmd(UniteNames.addable)

    @may_handle(UniteSelectAddAll)
    def select_add_all(self):
        self.unite_cmd(UniteNames.all_addable)

    @may_handle(UniteProjects)
    def projects(self):
        self.unite_cmd(UniteNames.projects)


class Plugin(ProteomeComponent):
    Transitions = UniteTransitions

    def __init__(self, *a, **kw):
        super().__init__(*a, **kw)
        self._unite_ready = False

    def prepare(self, msg):
        if (not self._unite_ready and isinstance(msg, UniteMessage)):
            self._setup_unite()

    def _setup_unite(self):
        addable = UniteSource(UniteNames.addable, UniteNames.addable_candidates, UniteNames.addable)
        all_addable = UniteSource(UniteNames.all_addable, UniteNames.all_addable_candidates,
                                  UniteNames.addable)
        add_action = Map(name='add', handler=UniteNames.add_project,
                         desc='add project')
        add_pro = UniteKind(UniteNames.addable, List(add_action))
        projects = UniteSource(UniteNames.projects, UniteNames.projects_candidates, UniteNames.project)
        delete_action = Map(name='delete', handler=UniteNames.delete_project,
                            desc='delete project')
        activate_action = Map(name='activate', handler=UniteNames.activate_project,
                              desc='activate project', is_selectable=0)
        project = UniteKind(UniteNames.project, List(activate_action, delete_action))
        addable.define(self.vim)
        all_addable.define(self.vim)
        add_pro.define(self.vim)
        projects.define(self.vim)
        project.define(self.vim)
        self._unite_ready = True

__all__ = ('Plugin')



from ribosome.machine import may_handle, message
from ribosome.nvim import Buffer

from amino import List, Empty, _
from amino.lazy import lazy

from proteome.state import ProteomeComponent, ProteomeTransitions
from proteome.ctags import Ctags
from proteome.plugins.core import Save, Added, BufEnter, StageIV

Gen = message('Gen', 'project')
GenAll = message('GenAll')
Kill = message('Kill')
CurrentBuffer = message('CurrentBuffer')


class CtagsTransitions(ProteomeTransitions):

    @property
    def ctags(self) -> Ctags:
        return self.machine.ctags  # type: ignore

    @property
    def _tags_file_name(self):
        return '.tags'

    @may_handle(StageIV)
    def stage_4(self):
        return GenAll(), CurrentBuffer()

    @may_handle(Save)
    def save(self):
        return GenAll()

    @may_handle(GenAll)
    def gen_all(self):
        if self.ctags.ready:
            return self.data.projects.projects\
                .filter(_.want_ctags)\
                .map(Gen)

    # TODO kill dangling procs
    @may_handle(Kill)
    def kill(self):
        pass

    @may_handle(BufEnter)
    def buf_enter(self):
        self.set_buffer_tags(List(self.msg.buffer))

    @may_handle(Added)
    def added(self: Added):
        return CurrentBuffer()

    @may_handle(CurrentBuffer)
    def buffer(self):
        self.set_buffer_tags(List(self.vim.buffer))

    def set_buffer_tags(self, bufs: List[Buffer]):
        files = self.data.all_projects.map(_.root / self._tags_file_name)
        bufs.foreach(lambda a: a.options.amend_l('tags', files))

    @may_handle(Gen)
    async def gen(self):
        await self.ctags.gen(self.msg.project)
        return Empty()


class Plugin(ProteomeComponent):
    Transitions = CtagsTransitions

    @lazy
    def ctags(self):
        return Ctags(self.vim)

__all__ = ('GenAll', 'Plugin')



import neovim

from amino import List, Map, __, _, Path

from ribosome import command, NvimStatePlugin, msg_command, json_msg_command
from ribosome.unite import mk_unite_candidates, mk_unite_action
from ribosome.unite.plugin import unite_plugin

from proteome.plugins.core import (
    AddByParams, Show, Create, SetProject, Next, Prev, StageI, Save, RemoveByIdent, BufEnter, StageII, StageIII,
    StageIV, CloneRepo
)
from proteome.plugins.history.messages import (HistoryPrev, HistoryNext, HistoryStatus, HistoryLog, HistoryBrowse,
                                               HistoryBrowseInput, HistorySwitch, HistoryPick, HistoryRevert,
                                               HistoryFileBrowse)
from proteome.main import Proteome
from proteome.nvim import NvimFacade
from proteome.logging import Logging
from proteome.plugins.unite import UniteSelectAdd, UniteSelectAddAll, UniteProjects, UniteNames

unite_candidates = mk_unite_candidates(UniteNames)
unite_action = mk_unite_action(UniteNames)


@unite_plugin('pro')
class ProteomeNvimPlugin(NvimStatePlugin, Logging, name='proteome', prefix='pro'):

    def __init__(self, vim: neovim.Nvim) -> None:
        super().__init__(NvimFacade(vim))
        self.initialized = False
        self.init_state()

    def init_state(self) -> None:
        config_path = self.vim.vars.ppath('config_path')\
            .get_or_else(Path('/dev/null'))
        bases = self.vim.vars.ppathl('base_dirs')\
            .get_or_else(List())\
            .map(Path)
        type_bases = self.vim.vars.pd('type_base_dirs')\
            .get_or_else(Map())\
            .keymap(lambda a: Path(a).expanduser())\
            .valmap(List.wrap)
        plugins = self.vim.vars.pl('plugins') | List()
        self.pro = Proteome(self.vim.proxy, Path(config_path), plugins, bases, type_bases)
        self.pro.start()
        self.pro.wait_for_running()

    def state(self):
        return self.pro

    def stage_1(self):
        self.pro.send(StageI())

    def stage_2(self):
        self.initialized = True
        self.pro.send(StageII().at(.7))

    def stage_3(self):
        self.pro.send(StageIII().at(.72))

    def stage_4(self):
        self.pro.send(StageIV().at(.74))

    @command()
    def pro_plug(self, plug_name, cmd_name, *args):
        self.pro.plug_command(plug_name, cmd_name, args)

    @msg_command(Create)
    def pro_create(self):
        pass

    addable = dict(complete='customlist,ProCompleteAddableProjects')
    projects = dict(complete='customlist,ProCompleteProjects')

    @json_msg_command(AddByParams, bang=True, **addable)
    def pro_add(self):
        pass

    @neovim.function('ProCompleteProjects', sync=True)
    def pro_complete_projects(self, args):
        lead, line, pos = args
        return self.pro.data.projects.idents.filter(__.startswith(lead))

    @neovim.function('ProCompleteAddableProjects', sync=True)
    def pro_complete_addable_projects(self, args):
        lead, line, pos = args
        return self.pro.data.addable.filter(__.startswith(lead))

    @msg_command(RemoveByIdent, **projects)
    def pro_remove(self):
        pass

    @msg_command(Show)
    def pro_show(self):
        pass

    @msg_command(SetProject, **projects)
    def pro_to(self):
        pass

    @msg_command(Next)
    def pro_next(self):
        pass

    @msg_command(Prev)
    def pro_prev(self):
        pass

    @msg_command(Save)
    def pro_save(self):
        pass

    # TODO start terminal at root dir
    # @msg_command(Term)
    # def pro_term(self):
        # pass

    @neovim.autocmd('BufEnter')
    def buf_enter(self):
        if self.initialized:
            self.pro.send(BufEnter(self.vim.buffer.proxy))

    @json_msg_command(CloneRepo)
    def pro_clone(self):
        pass

    @msg_command(HistoryPrev)
    def pro_history_prev(self):
        pass

    @msg_command(HistoryNext)
    def pro_history_next(self):
        pass

    @msg_command(HistoryStatus)
    def pro_history_status(self):
        pass

    @msg_command(HistoryLog)
    def pro_history_log(self):
        pass

    @msg_command(HistoryBrowse)
    def pro_history_browse(self):
        pass

    @msg_command(HistoryFileBrowse)
    def pro_history_file_browse(self):
        pass

    @msg_command(HistoryBrowseInput)
    def pro_history_browse_input(self):
        pass

    @msg_command(HistorySwitch)
    def pro_history_switch(self):
        pass

    @msg_command(HistoryPick)
    def pro_history_pick(self):
        pass

    @msg_command(HistoryRevert)
    def pro_history_revert(self):
        pass

    @msg_command(UniteSelectAdd)
    def pro_select_add(self):
        pass

    @msg_command(UniteSelectAddAll)
    def pro_select_add_all(self):
        pass

    @msg_command(UniteProjects)
    def projects(self):
        pass

    @unite_candidates('addable')
    def pro_unite_addable(self, args):
        return self.pro.data.main_addable

    @unite_candidates('all_addable')
    def pro_unite_all_addable(self, args):
        return self.pro.data.addable

    @unite_candidates('projects')
    def pro_unite_projects(self, args):
        return self.pro.data.all_projects / _.name

    @unite_action('add_project')
    def pro_unite_add_project(self, ident):
        return AddByParams(ident, options=Map())

    @unite_action('activate_project')
    def pro_unite_activate_project(self, ident):
        return SetProject(ident)

    @unite_action('delete_project')
    def pro_unite_delete_project(self, ident):
        return RemoveByIdent(ident)

__all__ = ('ProteomeNvimPlugin',)




from ribosome.machine import handle
from ribosome.machine.transition import Error

from amino import List, L, _

from proteome.state import ProteomeComponent, ProteomeTransitions
from proteome.plugins.core import StageII, StageIII


class Plugin(ProteomeComponent):

    class Transitions(ProteomeTransitions):

        @property
        def _project_dir(self):
            return self.vim.vars.p('config_project_dir')\
                .get_or_else('project')

        @property
        def _project_after_dir(self):
            return self.vim.vars.p('config_project_after_dir')\
                .get_or_else('project_after')

        def _runtime(self, project, base):
            ''' first runtimes all project type files, like
            `project/type.vim`, then the specific project file, like
            `project/type/name.vim`.
            '''
            def run(path_suf):
                path = '{}/{}'.format(base, path_suf)
                err = 'error sourcing {}.vim: {{}}'.format(path)
                return (
                    self.vim.runtime(path)
                    .cata(L(err.format)(_) >> List, lambda a: List())
                )
            return (
                project.all_types.flat_map(run) +
                (project.tpe.map(_ + '/' + project.name).map(run) | List()) +
                run('all/*')
            ).map(Error)

        def _runtime_before(self, project):
            return self._runtime(project, self._project_dir)

        def _runtime_after(self, project):
            return self._runtime(project, self._project_after_dir)

        @handle(StageII)
        def before(self):
            return self.data.current.map(self._runtime_before)

        @handle(StageIII)
        def stage_3(self):
            return self.data.current.map(self._runtime_after)

__all__ = ('Plugin',)



from proteome.project import Project
from ribosome import NvimProcessExecutor, Job


class Ctags(NvimProcessExecutor):

    def gen(self, project: Project):
        langs = ','.join(project.ctags_langs)
        tag_file = project.tag_file
        args = [
            '-R',
            '--languages={}'.format(langs),
            '-f',
            str(tag_file),
            str(project.root)
        ]
        job = Job(
            client=project.job_client,
            exe='ctags',
            args=args,
            loop=self.loop)
        return self.run(job)

__all__ = ('Ctags',)



import abc

from ribosome.machine import Message
from ribosome import NvimFacade

from amino import List, Map, _, L, __

from proteome.logging import Logging


class UniteMessage(Message, varargs='unite_args'):
    pass


class UniteSelectAdd(UniteMessage):
    pass


class UniteSelectAddAll(UniteMessage):
    pass


class UniteProjects(UniteMessage):
    pass


class UniteEntity(Logging, metaclass=abc.ABCMeta):
    ''' set up sources and kinds dynamically.
    The python call API cannot be used, as funcrefs cannot be
    serialized.
    The callback functions must be called once so that exists()
    can see them, otherwise Unite refuses to work.
    They must be called a/sync according to their definition,
    otherwise it will silently deadlock!
    '''

    def __init__(self, name: str) -> None:
        self.name = name

    @abc.abstractproperty
    def tpe(self) -> str:
        ...

    @abc.abstractproperty
    def data(self) -> str:
        ...

    @property
    def _func_defs_sync(self) -> List[str]:
        return List()

    @property
    def _func_defs_async(self) -> List[str]:
        return List()

    def _force_function_defs(self, vim: NvimFacade) -> None:
        force = lambda c, n: c('silent call {}()'.format(n))
        self._func_defs_sync.foreach(L(force)(vim.cmd_sync, _))
        self._func_defs_async.foreach(L(force)(vim.cmd, _))

    def define(self, vim: NvimFacade) -> None:
        self._force_function_defs(vim)
        vim.cmd('call unite#define_{}({})'.format(self.tpe, self.data))


class UniteSource(UniteEntity):
    _templ = '''
    {{
        'name': '{}',
        'gather_candidates': function('{}'),
        'default_kind': '{}',
    }}
    '''.replace('\n', '')

    def __init__(self, name: str, source: str, kind: str) -> None:
        super().__init__(name)
        self.source = source
        self.kind = kind

    @property
    def tpe(self) -> str:
        return 'source'

    @property
    def _func_defs_sync(self) -> List[str]:
        return List(self.source)

    @property
    def data(self) -> str:
        return self._templ.format(self.name, self.source, self.kind)


class UniteKind(UniteEntity):
    _templ = '''
    {{
        'name': '{name}',
        'default_action': '{default}',
        'parents': [],
        'action_table': {{
            {actions}
        }}
    }}
    '''.replace('\n', '')

    _action_templ = '''
    '{name}': {{
        'func': function('{handler}'),
        'description': '{desc}',
        'is_selectable': '{is_selectable}',
    }}
    '''.replace('\n', '')

    _defaults = Map(is_selectable=1)

    @property
    def tpe(self) -> str:
        return 'kind'

    def __init__(self, name: str, actions: List[Map]) -> None:
        super().__init__(name)
        self.actions = actions / self._defaults.merge
        self.default = actions.head / __['name'] | 'none'

    @property
    def _func_defs_async(self) -> List[str]:
        return self.actions / __['handler']

    def _action(self, params: dict) -> str:
        return self._action_templ.format(**params)

    @property
    def data(self) -> str:
        actions = self.actions.map(self._action).mk_string(', ')
        return self._templ.format(name=self.name, actions=actions, default=self.default)


class UniteNames():
    addable_candidates = '_proteome_unite_addable'
    all_addable_candidates = '_proteome_unite_all_addable'
    projects_candidates = '_proteome_unite_projects'

    add_project = '_proteome_unite_add_project'
    delete_project = '_proteome_unite_delete_project'
    activate_project = '_proteome_unite_activate_project'

    addable = 'proteome_addable'
    all_addable = 'proteome_all_addable'
    projects = 'proteome_projects'
    project = 'proteome_project'

__all__ = ('UniteNames', 'UniteSelectAdd', 'UniteSelectAddAll', 'UniteProjects')



from ribosome import NvimProcessExecutor, Job

from proteome.project import Project
from proteome.git.repo import CommitInfo


class Git(NvimProcessExecutor):

    def pre_args(self, project: Project):
        return []

    def command(self, client, git_args, name, *cmd_args):
        args = list(map(str, git_args + [name] + list(cmd_args)))
        self.log.debug('running git {}'.format(' '.join(args)))
        job = Job(
            client=client,
            exe='git',
            args=args,
            loop=self.loop,
        )
        return self.run(job)

    def project_command(self, project: Project, name: str, *cmd_args):
        return self.command(project.job_client, self.pre_args(project), name,
                            *cmd_args)

    def revert(self, project: Project, commit: CommitInfo):
        return self.project_command(project, 'revert', '-n', commit.hex)

    def revert_abort(self, project: Project):
        return self.project_command(project, 'revert', '--abort')

    def clone(self, client, url, location):
        self.log.info('Cloning {}...'.format(url))
        return self.command(client, [], 'clone', url, location)

    def add_all(self, project: Project):
        return self.project_command(project, 'add', '-A', '.')

    def commit(self, project: Project, msg: str):
        return self.project_command(project, 'commit', '-m', msg)

__all__ = ('Git',)
